/* 
 * tclxmlStubInit.c --
 */

#include "tclxml.h"

/*
 * Remove macros that will interfere with the definitions below.
 */


/*
 * WARNING: The contents of this file is automatically generated by the
 * genStubs.tcl script. Any modifications to the function declarations
 * below should be made in the tclxml.decls script.
 */

/* !BEGIN!: Do not edit below this line. */

TclxmlStubs tclxmlStubs = {
    TCL_STUB_MAGIC,
    NULL,
    Tclxml_Init, /* 0 */
    Tclxml_SafeInit, /* 1 */
    TclXML_RegisterXMLParser, /* 2 */
    TclXML_RegisterElementStartProc, /* 3 */
    TclXML_RegisterElementEndProc, /* 4 */
    TclXML_RegisterCharacterDataProc, /* 5 */
    TclXML_RegisterPIProc, /* 6 */
    TclXML_RegisterDefaultProc, /* 7 */
    TclXML_RegisterUnparsedProc, /* 8 */
    TclXML_RegisterNotationDeclProc, /* 9 */
    TclXML_RegisterEntityProc, /* 10 */
    TclXML_RegisterUnknownEncodingProc, /* 11 */
    TclXML_RegisterCommentProc, /* 12 */
    TclXML_RegisterNotStandaloneProc, /* 13 */
    TclXML_RegisterElementDeclProc, /* 14 */
    TclXML_RegisterAttListDeclProc, /* 15 */
    TclXML_RegisterStartDoctypeDeclProc, /* 16 */
    TclXML_RegisterEndDoctypeDeclProc, /* 17 */
    TclXML_ElementStartHandler, /* 18 */
    TclXML_ElementEndHandler, /* 19 */
    TclXML_CharacterDataHandler, /* 20 */
    TclXML_ProcessingInstructionHandler, /* 21 */
    TclXML_ExternalEntityRefHandler, /* 22 */
    TclXML_DefaultHandler, /* 23 */
    TclXML_UnparsedDeclHandler, /* 24 */
    TclXML_NotationDeclHandler, /* 25 */
    TclXML_UnknownEncodingHandler, /* 26 */
    TclXML_CommentHandler, /* 27 */
    TclXML_NotStandaloneHandler, /* 28 */
    NULL, /* 29 */
    NULL, /* 30 */
    TclXML_ElementDeclHandler, /* 31 */
    TclXML_AttlistDeclHandler, /* 32 */
    TclXML_StartDoctypeDeclHandler, /* 33 */
    TclXML_EndDoctypeDeclHandler, /* 34 */
};

/* !END!: Do not edit above this line. */
