// RCS: @(#) $Id: tclxml.rc,v 1.1 2002/09/12 21:15:56 balls Exp $
//
// Version resource script.
//

#include <winver.h>
#include "tclxml.h"

//
// build-up the name suffix that defines the type of build this is.
//
#ifdef DEBUG
#define SUFFIX_DEBUG	    "d"
#else
#define SUFFIX_DEBUG	    ""
#endif

#define SUFFIX		    SUFFIX_DEBUG


VS_VERSION_INFO VERSIONINFO
 FILEVERSION	TCLXML_MAJOR_VERSION ,TCLXML_MINOR_VERSION ,TCLXML_RELEASE_SERIAL, 0
 PRODUCTVERSION	TCLXML_MAJOR_VERSION ,TCLXML_MINOR_VERSION ,TCLXML_RELEASE_SERIAL, 0
 FILEFLAGSMASK	0x3fL
#ifdef DEBUG
 FILEFLAGS 	VS_FF_DEBUG
#else
 FILEFLAGS 	0x0L
#endif
 FILEOS 	VOS__WINDOWS32
 FILETYPE 	VFT_DLL
 FILESUBTYPE	0x0L
BEGIN
	BLOCK "StringFileInfo"
	BEGIN
		BLOCK "040904b0"
		BEGIN
			VALUE "FileDescription", "TCLXML extension for Tcl\0"
			VALUE "Authors", "\0"
			VALUE "OriginalFilename", "tclxml" STRINGIFY(JOIN(TCLXML_MAJOR_VERSION,TCLXML_MINOR_VERSION)) SUFFIX ".dll\0"
			VALUE "CompanyName", "\0"
			VALUE "FileVersion", TCLXML_PATCH_LEVEL
			VALUE "LegalCopyright", "Copyright \251 1993-2001\0"
			VALUE "ProductName", "tclxml " TCLXML_VERSION " for Windows\0"
			VALUE "ProductVersion", TCLXML_PATCH_LEVEL
		END
	END
	BLOCK "VarFileInfo"
	BEGIN
		VALUE "Translation", 0x409, 1200
	END
END




